defmodule Chronoscope.NTS.KeyEstablishment do
  import Bitwise

  @next_protocol_negotiation <<0x80, 0x01, 0x00, 0x02, 0x00, 0x00>>
  @aead_algorithm_negotiation <<0x80, 0x04, 0x00, 0x04, 0x00, 0x1E, 0x00, 0x0F>>
  @end_of_message <<0x80, 0x00, 0x00, 0x00>>

  @aead_alogorithms %{
    15 => "AEAD_AES_SIV_CMAC_256",
    30 => "AEAD_AES_128_GCM_SIV"
  }

  @next_protocols %{
    0 => "NTPv4"
  }

  @errors %{
    0 => "Unrecognized Critical Record",
    1 => "Bad Request",
    2 => "Internal Server Error"
  }

  def request() do
    @next_protocol_negotiation <> @aead_algorithm_negotiation <> @end_of_message
  end

  def parse_response(response) do
    {:ok, do_parse_response(response, %{})}
  end

  defp do_parse_response([], acc) do
    acc
  end

  # End of Message
  defp do_parse_response([0x80, 0x00, 0x00, 0x00 | _rest], acc) do
    acc
  end

  # NTS Next Protocol Negotiation
  defp do_parse_response([0x80, 0x01, length_high, length_low | rest], acc) do
    length = combine_octets(length_high, length_low)
    {next_protocols, remaining} = Enum.split(rest, length)

    do_parse_response(
      remaining,
      Map.put(acc, :next_protocols, parse_next_protocol_list(next_protocols))
    )
  end

  # Error
  defp do_parse_response([0x80, 0x02, 0x00, 0x02, error_high, error_low | rest], acc) do
    error = combine_octets(error_high, error_low)

    do_parse_response(
      rest,
      Map.put(acc, :error, Map.get(@errors, error, to_string(error)))
    )
  end

  # Warning
  defp do_parse_response([0x80, 0x03, 0x00, 0x02, warning_high, warning_low | rest], acc) do
    warning = combine_octets(warning_high, warning_low)

    do_parse_response(rest, Map.put(acc, :warning, to_string(warning)))
  end

  # AEAD Algorithm Negotiation
  defp do_parse_response([first, 0x04, length_high, length_low | rest], acc)
       when first == 0x00 or first == 0x80 do
    length = combine_octets(length_high, length_low)
    {aead_algorithms, remaining} = Enum.split(rest, length)

    do_parse_response(
      remaining,
      Map.put(acc, :aead_algorithms, parse_aead_algorithm_list(aead_algorithms))
    )
  end

  # New Cookie for NTPv4
  defp do_parse_response([first, 0x05, length_high, length_low | rest], acc) when first == 0x00 or first == 0x80 do
    length = combine_octets(length_high, length_low)
    {cookie, remaining} = Enum.split(rest, length)

    do_parse_response(
      remaining,
      acc
      |> Map.update(:cookies, [cookie], &[cookie | &1])
      |> Map.update(:cookie_length, length, &max(&1, length))
    )
  end

  # NTPv4 Server Negotiation
  defp do_parse_response([first, 0x06, length_high, length_low | rest], acc) when first == 0x00 or first == 0x80 do
    length = combine_octets(length_high, length_low)
    {server, remaining} = Enum.split(rest, length)

    do_parse_response(remaining, Map.put(acc, :server, to_string(server)))
  end

  # NTPv4 Port Negotiation
  defp do_parse_response([first, 0x07, 0x00, 0x02, port_high, port_low | rest], acc) when first == 0x00 or first == 0x80 do
    port = combine_octets(port_high, port_low)

    do_parse_response(rest, Map.put(acc, :port, port))
  end

  defp do_parse_response([_, _, length_high, length_low | rest], acc) do
    length = combine_octets(length_high, length_low)
    {_, remaining} = Enum.split(rest, length)

    do_parse_response(remaining, acc)
  end

  defp parse_next_protocol_list(next_protocols) do
    do_parse_next_protocol_list(next_protocols, [])
  end

  defp do_parse_next_protocol_list([], acc) do
    acc
  end

  defp do_parse_next_protocol_list([high, low | rest], acc) do
    @next_protocols
    |> Map.get(combine_octets(high, low), "UNASSIGNED")
    |> then(&do_parse_next_protocol_list(rest, [&1 | acc]))
  end

  defp parse_aead_algorithm_list(aead_algorithms) do
    do_parse_aead_algorithm_list(aead_algorithms, [])
  end

  defp do_parse_aead_algorithm_list([], acc) do
    acc
  end

  defp do_parse_aead_algorithm_list([high, low | rest], acc) do
    @aead_alogorithms
    |> Map.get(combine_octets(high, low), "UNKNOWN")
    |> then(&do_parse_aead_algorithm_list(rest, [&1 | acc]))
  end

  defp combine_octets(high, low) do
    high <<< 8 ||| low
  end
end
