defmodule Chronoscope.NTS.KeyEstablishmentTest do
  use ExUnit.Case

  alias Chronoscope.NTS.KeyEstablishment

  test "builds request" do
    assert KeyEstablishment.request() ==
             <<0x80, 0x01, 0x00, 0x02, 0x00, 0x00, 0x80, 0x04, 0x00, 0x04, 0x00, 0x1E, 0x00, 0x0F, 0x80, 0x00, 0x00, 0x00>>
  end

  test "handles empty response" do
    assert KeyEstablishment.parse_response([]) == {:ok, %{}}
  end

  test "ignores unkown record" do
    assert KeyEstablishment.parse_response([0x80, 0x21, 0x00, 0x04, 0x00, 0x01, 0x02, 0x03]) == {:ok, %{}}
  end

  test "handles end of message record" do
    assert KeyEstablishment.parse_response([0x80, 0x00, 0x00, 0x00]) == {:ok, %{}}
  end

  test "handles next protocol negotiation record" do
    assert KeyEstablishment.parse_response([0x80, 0x01, 0x00, 0x02, 0x00, 0x00]) == {:ok, %{next_protocols: ["NTPv4"]}}
  end

  test "does not handle next protocol negotiation record without critical bit" do
    assert KeyEstablishment.parse_response([0x00, 0x01, 0x00, 0x02, 0x00, 0x00]) == {:ok, %{}}
  end

  test "handles empty next protocols" do
    assert KeyEstablishment.parse_response([0x80, 0x01, 0x00, 0x00]) == {:ok, %{next_protocols: []}}
  end

  test "handles multiple next protocols" do
    assert KeyEstablishment.parse_response([0x80, 0x01, 0x00, 0x06, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00]) ==
             {:ok, %{next_protocols: ["NTPv4", "UNASSIGNED", "NTPv4"]}}
  end

  test "handles aead algorithm negotiation record" do
    assert KeyEstablishment.parse_response([0x80, 0x04, 0x00, 0x02, 0x00, 0x0F]) ==
             {:ok, %{aead_algorithms: ["AEAD_AES_SIV_CMAC_256"]}}
  end

  test "handles aead algorithm negotiation record without critical bit" do
    assert KeyEstablishment.parse_response([0x00, 0x04, 0x00, 0x02, 0x00, 0x1E]) ==
             {:ok, %{aead_algorithms: ["AEAD_AES_128_GCM_SIV"]}}
  end

  test "handles empty aead algorithms" do
    assert KeyEstablishment.parse_response([0x80, 0x04, 0x00, 0x00]) == {:ok, %{aead_algorithms: []}}
  end

  test "handles multiple aead algorithms" do
    assert KeyEstablishment.parse_response([0x80, 0x04, 0x00, 0x06, 0x00, 0x1E, 0x00, 0x01, 0x00, 0x0F]) ==
             {:ok, %{aead_algorithms: ["AEAD_AES_SIV_CMAC_256", "UNKNOWN", "AEAD_AES_128_GCM_SIV"]}}
  end

  test "handles error record" do
    assert KeyEstablishment.parse_response([0x80, 0x02, 0x00, 0x02, 0x00, 0x01]) == {:ok, %{error: "Bad Request"}}
  end

  test "handles unknown error record" do
    assert KeyEstablishment.parse_response([0x80, 0x02, 0x00, 0x02, 0x00, 0x99]) == {:ok, %{error: "153"}}
  end

  test "does not handle error record without critical bit" do
    assert KeyEstablishment.parse_response([0x00, 0x02, 0x00, 0x02, 0x00, 0x01]) == {:ok, %{}}
  end

  test "handles warning record" do
    assert KeyEstablishment.parse_response([0x80, 0x03, 0x00, 0x02, 0x00, 0x01]) == {:ok, %{warning: "1"}}
  end

  test "does not handle warning record without critical bit" do
    assert KeyEstablishment.parse_response([0x00, 0x03, 0x00, 0x02, 0x00, 0x01]) == {:ok, %{}}
  end

  test "handles server record" do
    assert KeyEstablishment.parse_response([0x80, 0x06, 0x00, 0x09, ?1, ?2, ?7, ?., ?0, ?., ?0, ?., ?1]) ==
             {:ok, %{server: "127.0.0.1"}}
  end

  test "handles server record without critical bit" do
    assert KeyEstablishment.parse_response([0x00, 0x06, 0x00, 0x09, ?1, ?2, ?7, ?., ?0, ?., ?0, ?., ?1]) ==
             {:ok, %{server: "127.0.0.1"}}
  end

  test "handles port record" do
    assert KeyEstablishment.parse_response([0x80, 0x07, 0x00, 0x02, 0x04, 0xCE]) == {:ok, %{port: 1230}}
  end

  test "handles port record without critical bit" do
    assert KeyEstablishment.parse_response([0x00, 0x07, 0x00, 0x02, 0x04, 0xCE]) == {:ok, %{port: 1230}}
  end

  test "handles cookie record" do
    assert KeyEstablishment.parse_response([0x80, 0x05, 0x00, 0x0E, ?c, ?h, ?o, ?c, ?o, ?l, ?a, ?t, ?e, ?_, ?c, ?h, ?i, ?p]) ==
             {:ok, %{cookies: ['chocolate_chip'], cookie_length: 14}}
  end

  test "handles cookie record without critical bit" do
    assert KeyEstablishment.parse_response([0x00, 0x05, 0x00, 0x0E, ?c, ?h, ?o, ?c, ?o, ?l, ?a, ?t, ?e, ?_, ?c, ?h, ?i, ?p]) ==
             {:ok, %{cookies: ['chocolate_chip'], cookie_length: 14}}
  end

  test "sets cookie length to longest cookie" do
    assert KeyEstablishment.parse_response(
             [0x80, 0x05, 0x00, 0x01, ?c] ++
               [0x80, 0x05, 0x00, 0x03, ?c, ?c, ?c] ++
               [0x80, 0x05, 0x00, 0x01, ?c]
           ) ==
             {:ok, %{cookies: ['c', 'ccc', 'c'], cookie_length: 3}}
  end

  test "handles full response" do
    assert KeyEstablishment.parse_response(
             [0x80, 0x01, 0x00, 0x02, 0x00, 0x00] ++
               [0x80, 0x04, 0x00, 0x06, 0x00, 0x1E, 0x00, 0x01, 0x00, 0x0F] ++
               [0x80, 0x05, 0x00, 0x01, ?c] ++
               [0x00, 0x21, 0x00, 0x04, 0x00, 0x01, 0x02, 0x03] ++
               [0x80, 0x06, 0x00, 0x09, ?1, ?2, ?7, ?., ?0, ?., ?0, ?., ?1] ++
               [0x80, 0x07, 0x00, 0x02, 0x04, 0xCE] ++
               [0x80, 0x00, 0x00, 0x00]
           ) ==
             {:ok,
              %{
                cookies: ['c'],
                aead_algorithms: ["AEAD_AES_SIV_CMAC_256", "UNKNOWN", "AEAD_AES_128_GCM_SIV"],
                cookie_length: 1,
                next_protocols: ["NTPv4"],
                port: 1230,
                server: "127.0.0.1"
              }}
  end

  test "doesn't read past end of message record" do
    assert KeyEstablishment.parse_response(
             [0x80, 0x01, 0x00, 0x02, 0x00, 0x00] ++
               [0x00, 0x21, 0x00, 0x04, 0x00, 0x01, 0x02, 0x03] ++
               [0x80, 0x04, 0x00, 0x02, 0x00, 0x1E] ++
               [0x80, 0x00, 0x00, 0x00] ++
               [0x80, 0x05, 0x00, 0x01, ?c] ++
               [0x80, 0x06, 0x00, 0x09, ?1, ?2, ?7, ?., ?0, ?., ?0, ?., ?1] ++
               [0x80, 0x07, 0x00, 0x02, 0x04, 0xCE]
           ) ==
             {:ok,
              %{
                aead_algorithms: ["AEAD_AES_128_GCM_SIV"],
                next_protocols: ["NTPv4"]
              }}
  end
end
